@@ -20,7 +20,9 @@
 #--------------------------------------------------------------------------#
 
 
+import collections
 import glib
+import operator
 
 
 __all__ = ('TreeNode', 'DefaultTreeNode', 'LazyDefaultTreeNode')
@@ -37,7 +39,7 @@ class TreeNode (object):
 
 
     @property
-    def parent_node (self): raise NotImplemented
+    def parent_node (self): raise NotImplementedError
 
     # Default implementation does nothing: if you have other way of deriving parent node
     # for the property, you can do so without implementing this method.
@@ -46,7 +48,7 @@ class TreeNode (object):
 
 
     @property
-    def next_node (self): raise NotImplemented
+    def next_node (self): raise NotImplementedError
 
     # Default implementation does nothing: if you have other way of deriving next node for
     # the property, you can do so without implementing this method.
@@ -87,10 +89,10 @@ class TreeNode (object):
         return num_children
 
     @property
-    def first_child_node (self): raise NotImplemented
+    def first_child_node (self): raise NotImplementedError
 
-    # Default implementation does nothing: if you have other way of deriving parent node
-    # for the property, you can do so without implementing this method.
+    # Default implementation does nothing: if you have other way of deriving first child
+    # node for the property, you can do so without implementing this method.
     def _set_first_child_node (self, first_child_node):
         pass
 
@@ -162,10 +164,30 @@ class TreeNode (object):
 
 
     def delete_child_node (self, child):
-        raise NotImplemented
+        raise NotImplementedError
 
     def insert_child_node (self, child, before = None):
-        raise NotImplemented
+        raise NotImplementedError
+
+
+    @classmethod
+    def create_tree (cls, data):
+        if not data:
+            return None
+
+        data = tuple (data)
+        if not data:
+            return None
+
+        node            = cls ()
+        node.row_object = data[0]
+
+        for subdata in data[1:]:
+            child = cls.create_tree (subdata)
+            if child is not None:
+                node.insert_child_node (child)
+
+        return node
 
 
     def hierarchic_copy (self):
@@ -181,6 +203,114 @@ class TreeNode (object):
         return copy
 
 
+    @property
+    def is_draggable (self):
+        try:
+            return self.row_object.is_draggable
+        except AttributeError:
+            return True
+
+    def accepts_dragged_child_node (self, child_node):
+        try:
+            row_object_method = self.row_object.accepts_dragged_child_node
+        except AttributeError:
+            return True
+
+        return row_object_method (child_node)
+
+    def accepts_parent_node_after_drag (self, parent_node):
+        try:
+            row_object_method = self.row_object.accepts_parent_node_after_drag
+        except AttributeError:
+            return True
+
+        return row_object_method (parent_node)
+
+
+    # Traversal algorithms below should all check 'has_child_nodes_currently' before
+    # querying 'first_child_node'.  Otherwise, trees of lazy nodes would become fully
+    # loaded on traversal and we don't need that.
+
+    # This implementation uses no recursion and constant memory.
+    def depth_first_order (self, include_self = True):
+        if include_self:
+            yield self
+
+        if not self.has_child_nodes_currently:
+            return
+
+        node = self.first_child_node
+
+        while True:
+            yield node
+
+            if node.has_child_nodes_currently:
+                node = node.first_child_node
+                continue
+
+            while True:
+                if node is self:
+                    return
+
+                next = node.next_node
+                if next is not None:
+                    node = next
+                    break
+                else:
+                    node = node.parent_node
+
+
+    # Very similar to the above algorithm.
+    def traversal_postorder (self, include_self = True):
+        node = self
+
+        while True:
+            if node.has_child_nodes_currently:
+                node = node.first_child_node
+                continue
+
+            while True:
+                if node is self:
+                    if include_self:
+                        yield self
+
+                    return
+
+                yield node
+
+                next = node.next_node
+                if next is not None:
+                    node = next
+                    break
+                else:
+                    node = node.parent_node
+
+
+    # This implementation uses no recursion and O(K) memory, where K is maximum number of
+    # non-leaf nodes in a layer.
+    def breadth_first_order (self, include_self = True):
+        # Use special case for 'self', to avoid inspecting its next node.
+        if include_self:
+            yield self
+
+        if not self.has_child_nodes_currently:
+            return
+
+        queue = collections.deque ()
+        queue.append (self.first_child_node)
+
+        while queue:
+            node = queue.popleft ()
+
+            while node is not None:
+                yield node
+
+                if node.has_child_nodes_currently:
+                    queue.append (node.first_child_node)
+
+                node = node.next_node
+
+
     def __repr__(self):
         return '<%s.%s at 0x%x: %r>' % (type (self).__module__, type (self).__name__, id (self), self.row_object)
 
@@ -297,12 +427,58 @@ class _DefaultTreeNodeChildren (list):
         model, path = self._get_tree_model_and_node_path ()
         self._do_insert (model, path, len (self), (child,))
 
+    def extend (self, nodes):
+        model, path = self._get_tree_model_and_node_path ()
+        self._do_insert (model, path, len (self), nodes)
+
     def insert (self, index, child):
         model, path = self._get_tree_model_and_node_path ()
-        self._do_insert (model, path, index, (child,))
+        self._do_insert (model, path, self._get_insertion_point (index), (child,))
 
     def remove (self, value):
-        del self[self.index (value)]
+        model, path = self._get_tree_model_and_node_path ()
+        self._do_remove_at (model, path, (self.index (value),))
+
+    def pop (self, index = -1):
+        index = self._get_existing_child_index (index)
+        child = self[index]
+
+        model, path = self._get_tree_model_and_node_path ()
+        self._do_remove_at (model, path, (index,))
+
+        return child
+
+    def reverse (self):
+        raise NotImplementedError
+
+    def sort (self, cmp = None, key = None, reverse = False):
+        raise NotImplementedError
+
+
+    def _get_existing_child_index (self, index):
+        integer_index = operator.index (index)
+        if integer_index < 0:
+            integer_index += len (self)
+            if integer_index >= 0:
+                return integer_index
+        elif integer_index < len (self):
+            return integer_index
+
+        raise self._create_index_error (index)
+
+    def _get_insertion_point (self, index):
+        index = operator.index (index)
+        size  = len (self)
+
+        if index < 0:
+            index += size
+
+        return max (0, min (index, size))
+
+    def _create_index_error (self, index):
+        return IndexError ("invalid child node index %s: "
+                           "only integers in range %d..%d (inclusive) are valid)"
+                           % (index, -len (self), len (self) - 1))
 
 
     # 'indices' *must* be between 0 and len (self) and *must* be in reverse order (largest
@@ -341,7 +517,17 @@ class _DefaultTreeNodeChildren (list):
 
         return num_removed
 
-    def _do_insert (self, model, path, index, nodes, after = False):
+    def _do_insert (self, model, path, index, nodes):
+        nodes = tuple (nodes)
+
+        for child in nodes:
+            if not isinstance (child, TreeNode):
+                raise TypeError ("can only use TreeNode objects as child nodes")
+            if child.parent_node is not None:
+                raise ValueError ("new child nodes cannot already have a parent node")
+            if child._do_get_tree_model () is not None:
+                raise ValueError ("root node of a model's tree cannot become a child node")
+
         current_child = (self[index] if index < len (self) else None)
 
         for child in nodes:
@@ -382,7 +568,7 @@ class _DefaultTreeNodeChildren (list):
         model, path = self._get_tree_model_and_node_path ()
 
         if not isinstance (index, slice):
-            self._do_remove_at (model, path, (index,))
+            self._do_remove_at (model, path, (self._get_existing_child_index (index),))
         else:
             indices = range (index.start, min (index.stop, len (self)), index.step or 1)
             indices.reverse ()
@@ -392,6 +578,7 @@ class _DefaultTreeNodeChildren (list):
         model, path = self._get_tree_model_and_node_path ()
 
         if not isinstance (index, slice):
+            index = self._get_existing_child_index (index)
             self._do_remove_at (model, path, (index,))
             self._do_insert    (model, path, index, (node,))
             return
@@ -414,15 +601,14 @@ class _DefaultTreeNodeChildren (list):
                                   % (len (new_children), len (indices)))
 
 
-    if '__getslice__' in dir (list):
-        def __getslice__(self, start, stop):
-            self.__getitem__(slice (start, stop))
+    # These methods are deprecated, but we need to override them as long as underlying
+    # Python 'list' class has them.
 
-    if '__setslice__' in dir (list):
+    if hasattr (list, '__setslice__'):
         def __setslice__(self, start, stop, sequence):
             self.__setitem__(slice (start, stop), sequence)
 
-    if '__delslice__' in dir (list):
+    if hasattr (list, '__delslice__'):
         def __delslice__(self, start, stop):
             self.__delitem__(slice (start, stop))
 
@@ -441,13 +627,24 @@ class LazyDefaultTreeNode (DefaultTreeNode):
         self.__load_state = (0 if not child_nodes else 2)
 
 
-    # Default implementation actually just calls lazy_load().
     def lazy_has_children (self):
-        self._do_lazy_load ()
-        return self.has_child_nodes
+        try:
+            row_object_method = self.row_object.lazy_has_children
+        except AttributeError:
+            # Just calls lazy_load() then (indirectly).
+            self._do_lazy_load ()
+            return self.has_child_nodes
+
+        return row_object_method ()
 
     def lazy_load (self):
-        raise NotImplemented
+        try:
+            row_object_method = self.row_object.lazy_load
+        except AttributeError:
+            raise NotImplementedError ("lazy_load() should be provided by row object "
+                                       "or overriden in subclass")
+
+        return row_object_method ()
 
 
     def _determine_if_has_children (self):
@@ -464,7 +661,24 @@ class LazyDefaultTreeNode (DefaultTreeNode):
     def _do_lazy_load (self):
         self.__load_state = 2
         child_nodes       = self.child_nodes
-        loaded_nodes      = tuple (self.lazy_load ())
+        loaded_nodes      = list (self.lazy_load ())
+        default_node_type = None
+
+        for index, value in enumerate (loaded_nodes):
+            if not isinstance (value, TreeNode):
+                # Silently wrap returned value (as row object) in a node, lazy or not.
+                if hasattr (value, 'lazy_load'):
+                    node = type (self) ()
+                else:
+                    if default_node_type is None:
+                        model             = self.get_tree_model ()
+                        default_node_type = (model.default_node_type
+                                             if model is not None else DefaultTreeNode)
+
+                    node = default_node_type ()
+
+                node.row_object     = value
+                loaded_nodes[index] = node
 
         if not child_nodes or loaded_nodes:
             child_nodes[:] = loaded_nodes
