@@ -29,6 +29,16 @@ __all__ = ('RowObjectTreeModel',
            'TreeNode', 'DefaultTreeNode', 'LazyDefaultTreeNode', 'RowObjectTreeStore')
 
 
+# Should be available on 2.6 and up.
+_HAVE_DOTTED_ATTRGETTER_SUPPORT = False
+
+try:
+    if operator.attrgetter ('attrgetter.__call__') (operator) == operator.attrgetter.__call__:
+        _HAVE_DOTTED_ATTRGETTER_SUPPORT = True
+except:
+    pass
+
+
 
 class _Columns (tuple):
 
@@ -44,22 +54,27 @@ class _Columns (tuple):
 class RowObjectTreeModel (gtk.GenericTreeModel):
 
     def __init__(self, columns):
-        self.__accessors = []
-        self.__columns   = _Columns (columns)
+        self.__getters = []
+        self.__setters = []
+        self.__columns = _Columns (columns)
 
         for g_type, attribute in self.__columns:
             if isinstance (attribute, basestring):
-                accessor = operator.attrgetter (attribute)
+                getter = _attrgetter (attribute)
+                setter = _attrsetter (attribute)
             else:
                 assert callable (attribute)
-                accessor = attribute
+                getter = attribute
+                setter = _raising_attrsetter
 
-            self.__accessors.append (accessor)
+            self.__getters.append (getter)
+            self.__setters.append (setter)
 
         gtk.GenericTreeModel.__init__(self)
 
         self.__object_type          = None
         self.__foreign_drags_policy = None
+        self.__default_node_type    = DefaultTreeNode
 
 
     @property
@@ -71,8 +86,28 @@ class RowObjectTreeModel (gtk.GenericTreeModel):
         return False
 
 
+    # Strictly speaking, this property only makes sense for node-based model.
+
+    def _get_default_node_type (self):
+        return self.__default_node_type
+
+    def _set_default_node_type (self, default_node_type):
+        if not issubclass (default_node_type, TreeNode):
+            raise TypeError ("'default_node_type' must be a subtype of TreeNode")
+
+        self.__default_node_type = default_node_type
+
+    default_node_type = property (_get_default_node_type, _set_default_node_type)
+    del _get_default_node_type, _set_default_node_type
+
+
     def get_cell (self, row, attribute):
-        return self.get_value (self.get_iter (row), self.columns (attribute))
+        return self.get_value (self._do_get_iter (row), self.columns (attribute))
+
+    def set_cell (self, row, attribute, value):
+        iter = self._do_get_iter (row)
+        self.__setters[self.columns (attribute)] (self.get_row_object (iter), value)
+        self.note_changes (iter)
 
 
     def on_get_n_columns (self):
@@ -83,7 +118,7 @@ class RowObjectTreeModel (gtk.GenericTreeModel):
 
 
     def on_get_value (self, row, column):
-        return self.__accessors[column] (self._do_get_row_object (row))
+        return self.__getters[column] (self._do_get_row_object (row))
 
     def get_row_object (self, iter):
         return self._do_get_row_object (self.get_user_data (iter))
@@ -92,16 +127,16 @@ class RowObjectTreeModel (gtk.GenericTreeModel):
         return self._do_get_node (self.get_user_data (iter))
 
     def _do_get_row_object (self, row):
-        raise NotImplemented
+        raise NotImplementedError
 
     def _do_get_node (self, row):
         if self.node_based:
-            raise NotImplemented
+            raise NotImplementedError
         else:
             raise RuntimeError ("this model is not node-based")
 
     def _do_get_iter (self, row):
-        raise NotImplemented
+        raise NotImplementedError
 
 
     def note_changes (self, *rows):
@@ -174,16 +209,26 @@ class RowObjectTreeModel (gtk.GenericTreeModel):
 
     def _get_drag_data (self, selection_data):
         source_model, source_path = selection_data.tree_get_row_drag_data ()
-
-        if self.is_accepting_drags_from (source_model):
-            source_iter = source_model.get_iter (source_path)
-            if not self.node_based:
-                return source_model, source_path, source_model.get_row_object (source_iter)
-            else:
-                return source_model, source_path, source_model.get_node (source_iter)
-        else:
+        if not self.is_accepting_drags_from (source_model):
             return None, None, None
 
+        source_iter = source_model.get_iter (source_path)
+
+        if not self.node_based:
+            element = source_model.get_row_object (source_iter)
+        elif source_model.node_based:
+            element = source_model.get_node (source_iter)
+        else:
+            row = source_model.get_row_object (source_iter)
+
+            if isinstance (row, TreeNode) and row.row_object is row:
+                element = row
+            else:
+                element            = self.default_node_type ()
+                element.row_object = row
+
+        return source_model, source_path, element
+
 
 
 # IMPLEMENTATION NOTES
@@ -657,6 +702,7 @@ class RowObjectTreeStore (RowObjectTreeModel, gtk.TreeDragSource, gtk.TreeDragDe
     __gtype_name__ = 'RowObjectTreeStore'
 
     def __init__(self, columns, top_level_nodes = None):
+        # This postponed initialization allows to avoid unnecessary signal emissions.
         self.__root            = _RootTreeNode (None, top_level_nodes)
         self.__root.row_object = self
 
@@ -689,12 +735,6 @@ class RowObjectTreeStore (RowObjectTreeModel, gtk.TreeDragSource, gtk.TreeDragDe
         raise TypeError ("'%s' object is not iterable" % type (self).__name__)
 
 
-    def _create_index_error (self, index):
-        return IndexError ("invalid row index %s: "
-                           "only integers in range %d..%d (inclusive) and gtk.TreeIter are valid)"
-                           % (index, -len (self), len (self) - 1))
-
-
     def on_get_flags (self):
         return gtk.TREE_MODEL_ITERS_PERSIST
 
@@ -748,7 +788,8 @@ class RowObjectTreeStore (RowObjectTreeModel, gtk.TreeDragSource, gtk.TreeDragDe
     # Implementing gtk.TreeDragSource.
 
     def do_row_draggable (self, path):
-        return True
+        node = self.get_node (self.get_iter (path))
+        return node.is_draggable
 
     def do_drag_data_delete (self, path):
         node = self.get_node (self.get_iter (path))
@@ -763,34 +804,88 @@ class RowObjectTreeStore (RowObjectTreeModel, gtk.TreeDragSource, gtk.TreeDragDe
 
     def do_drag_data_received (self, dest_path, selection_data):
         source_model, source_path, node = self._get_drag_data (selection_data)
-
-        if source_model is not None:
-            node = node.hierarchic_copy ()
-
-            # This convoluted logic is meant to cope with drag-to-the-very-end situations.
-            if dest_path and dest_path[-1] > 0:
-                dest_path      = list (dest_path)
-                dest_path[-1] -= 1
-
-                dest_node = self.get_node (self.get_iter (tuple (dest_path)))
-                dest_node.parent_node.insert_child_node (node, dest_node.next_node)
-            else:
-                dest_node_parent = self.get_node (self.get_iter (dest_path)).parent_node
-                dest_node_parent.insert_child_node (node, dest_node_parent.first_child_node)
-
-            return True
-        else:
+        if source_model is None:
             return False
 
-    def do_row_drop_possible (self, dest_path, selection_data):
+        dest_parent, dest_before = self._get_drag_destination_nodes (dest_path)
+        dest_parent.insert_child_node (node.hierarchic_copy (), dest_before)
         return True
 
+    def do_row_drop_possible (self, dest_path, selection_data):
+        source_model, source_path, node = self._get_drag_data (selection_data)
+        if source_model is None:
+            return False
+
+        dest_parent, dest_before = self._get_drag_destination_nodes (dest_path)
+        return (dest_parent.accepts_dragged_child_node (node)
+                and node.accepts_parent_node_after_drag (dest_parent))
+
+    def _get_drag_destination_nodes (self, dest_path):
+        if dest_path[-1] > 0:
+            dest_path      = list (dest_path)
+            dest_path[-1] -= 1
+
+            dest_node = self.get_node (self.get_iter (tuple (dest_path)))
+            return dest_node.parent_node, dest_node.next_node
+
+        elif len (dest_path) > 1:
+            dest_node_parent = self.get_node (self.get_iter (dest_path[:-1]))
+            return dest_node_parent, dest_node_parent.first_child_node
+
+        else:
+            return self.root, self.root.first_child_node
+
+
 
 class _RootTreeNode (DefaultTreeNode):
 
     def _do_get_tree_model (self):
         return self.row_object
 
+    def _set_row_object (self, row_object):
+        if not hasattr (self, '_row_object') or self._row_object is None:
+            self._row_object = row_object
+        else:
+            raise RuntimeError ("cannot change row object of a model's tree root node")
+
+    row_object = property (DefaultTreeNode.row_object.fget, _set_row_object)
+    del _set_row_object
+
+
+
+def _attrgetter (attribute):
+    if '.' not in attribute or _HAVE_DOTTED_ATTRGETTER_SUPPORT:
+        return operator.attrgetter (attribute)
+    else:
+        name_parts = attribute.split ('.')
+        def dotted_attrgetter (object):
+            value = object
+            for part in name_parts:
+                value = getattr (value, part)
+            return value
+
+    return dotted_attrgetter
+
+
+def _attrsetter (attribute):
+    base, dot, final_attribute = attribute.rpartition ('.')
+
+    if base:
+        base_getter = _attrgetter (base)
+        def composite_setter (object, value):
+            setattr (base_getter (object), final_attribute, value)
+
+        return composite_setter
+
+    else:
+        def simple_setter (object, value):
+            setattr (object, final_attribute, value)
+
+        return simple_setter
+
+
+def _raising_attrsetter (attribute):
+    raise TypeError ("this attribute cannot be modified")
 
 
 # Local variables:
